/*
 * Interface to libfaac for aac encoding
 * Copyright (c) 2002 Gildas Bazin <gbazin@netcourrier.com>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file libavcodec/libfaac.c
 * Interface to libfaac for aac encoding.
 */

#include "avcodec.h"
#include <faac.h>

#ifndef _WIN32
#include <dlfcn.h>
#define LIBEXT "so"
#else
#include <windows.h>
#define LIBEXT "dll"
#define dlopen(fname, f) ((void *) LoadLibraryA(fname))
#define dlclose(handle) FreeLibrary((HMODULE) handle)
#define dlsym(handle, name) GetProcAddress((HMODULE) handle, name)
#endif

#ifndef FAACAPI
#define FAACAPI
#endif

typedef faacEncHandle (FAACAPI *imp_faacEncOpen)(unsigned long, unsigned int numChannels, unsigned long *, unsigned long *);
typedef int (FAACAPI *imp_faacEncClose)(faacEncHandle);
typedef int (FAACAPI *imp_faacEncEncode)(faacEncHandle , signed int *, unsigned int, unsigned char *, unsigned int);
typedef int (FAACAPI *imp_faacEncSetConfiguration)(faacEncHandle, faacEncConfigurationPtr);
typedef faacEncConfigurationPtr (FAACAPI *imp_faacEncGetCurrentConfiguration)(faacEncHandle);
typedef int (FAACAPI *imp_faacEncGetDecoderSpecificInfo)(faacEncHandle, unsigned char **, unsigned long *);

typedef struct _faacLib
{
    imp_faacEncOpen faacEncOpen;
    imp_faacEncClose faacEncClose;
    imp_faacEncEncode faacEncEncode;
    imp_faacEncSetConfiguration faacEncSetConfiguration;
    imp_faacEncGetCurrentConfiguration faacEncGetCurrentConfiguration;
    imp_faacEncGetDecoderSpecificInfo faacEncGetDecoderSpecificInfo;
} faacLib;

typedef struct FaacAudioContext {
    faacEncHandle faac_handle;
    void *hLib;
    faacLib pfn;
} FaacAudioContext;

#define DLSYM(x) \
    do \
    { \
        s->pfn.x = ( imp_##x ) dlsym(s->hLib, AV_STRINGIFY(x)); \
        if (!s->pfn.x ) \
        { \
            av_log(avctx, AV_LOG_ERROR, "Unable to find symbol " AV_STRINGIFY(x) " in dynamic libfaac." LIBEXT "\n"); \
            return -1; \
        } \
    } while (0)

static av_cold int Faac_encode_init(AVCodecContext *avctx)
{
    FaacAudioContext *s = avctx->priv_data;
    faacEncConfigurationPtr faac_cfg;
    unsigned long samples_input, max_bytes_output;

    if (!(s->hLib = dlopen("libfaac." LIBEXT, RTLD_NOW))) {
        av_log(avctx, AV_LOG_ERROR, "Unable to load libfaac." LIBEXT "\n");
        return -1;
    }
    
    DLSYM(faacEncOpen);
    DLSYM(faacEncClose);
    DLSYM(faacEncEncode);
    DLSYM(faacEncSetConfiguration);
    DLSYM(faacEncGetCurrentConfiguration);
    DLSYM(faacEncGetDecoderSpecificInfo);
    
    /* number of channels */
    if (avctx->channels < 1 || avctx->channels > 6) {
        av_log(avctx, AV_LOG_ERROR, "encoding %d channel(s) is not allowed\n", avctx->channels);
        return -1;
    }

    s->faac_handle = s->pfn.faacEncOpen(avctx->sample_rate,
                                         avctx->channels,
                                         &samples_input, &max_bytes_output);

    /* check faac version */
    faac_cfg = s->pfn.faacEncGetCurrentConfiguration(s->faac_handle);
    if (faac_cfg->version != FAAC_CFG_VERSION) {
        av_log(avctx, AV_LOG_ERROR, "wrong libfaac version (compiled for: %d, using %d)\n", FAAC_CFG_VERSION, faac_cfg->version);
        s->pfn.faacEncClose(s->faac_handle);
        return -1;
    }

    /* put the options in the configuration struct */
    switch(avctx->profile) {
        case FF_PROFILE_AAC_MAIN:
            faac_cfg->aacObjectType = MAIN;
            break;
        case FF_PROFILE_UNKNOWN:
        case FF_PROFILE_AAC_LOW:
            faac_cfg->aacObjectType = LOW;
            break;
        case FF_PROFILE_AAC_SSR:
            faac_cfg->aacObjectType = SSR;
            break;
        case FF_PROFILE_AAC_LTP:
            faac_cfg->aacObjectType = LTP;
            break;
        default:
            av_log(avctx, AV_LOG_ERROR, "invalid AAC profile\n");
            s->pfn.faacEncClose(s->faac_handle);
            return -1;
    }
    faac_cfg->mpegVersion = MPEG4;
    faac_cfg->useTns = 0;
    faac_cfg->allowMidside = 1;
    faac_cfg->bitRate = avctx->bit_rate / avctx->channels;
    faac_cfg->bandWidth = avctx->cutoff;
    if(avctx->flags & CODEC_FLAG_QSCALE) {
        faac_cfg->bitRate = 0;
        faac_cfg->quantqual = avctx->global_quality / FF_QP2LAMBDA;
    }
    faac_cfg->outputFormat = 1;
    faac_cfg->inputFormat = FAAC_INPUT_16BIT;

    avctx->frame_size = samples_input / avctx->channels;

    avctx->coded_frame= avcodec_alloc_frame();
    avctx->coded_frame->key_frame= 1;

    /* Set decoder specific info */
    avctx->extradata_size = 0;
    if (avctx->flags & CODEC_FLAG_GLOBAL_HEADER) {

        unsigned char *buffer = NULL;
        unsigned long decoder_specific_info_size;

        if (!s->pfn.faacEncGetDecoderSpecificInfo(s->faac_handle, &buffer,
                                                   &decoder_specific_info_size)) {
            avctx->extradata = av_malloc(decoder_specific_info_size + FF_INPUT_BUFFER_PADDING_SIZE);
            avctx->extradata_size = decoder_specific_info_size;
            memcpy(avctx->extradata, buffer, avctx->extradata_size);
            faac_cfg->outputFormat = 0;
        }
#undef free
        free(buffer);
#define free please_use_av_free
    }

    if (!s->pfn.faacEncSetConfiguration(s->faac_handle, faac_cfg)) {
        av_log(avctx, AV_LOG_ERROR, "libfaac doesn't support this output format!\n");
        return -1;
    }

    return 0;
}

static int Faac_encode_frame(AVCodecContext *avctx,
                             unsigned char *frame, int buf_size, void *data)
{
    FaacAudioContext *s = avctx->priv_data;
    int bytes_written;

    bytes_written = s->pfn.faacEncEncode(s->faac_handle,
                                         data,
                                         avctx->frame_size * avctx->channels,
                                         frame,
                                         buf_size);

    return bytes_written;
}

static av_cold int Faac_encode_close(AVCodecContext *avctx)
{
    FaacAudioContext *s = avctx->priv_data;

    av_freep(&avctx->coded_frame);
    av_freep(&avctx->extradata);

    if (s->hLib) {
        s->pfn.faacEncClose(s->faac_handle);
        dlclose(s->hLib);
    }
    return 0;
}

AVCodec libfaac_encoder = {
    "libfaac",
    AVMEDIA_TYPE_AUDIO,
    CODEC_ID_AAC,
    sizeof(FaacAudioContext),
    Faac_encode_init,
    Faac_encode_frame,
    Faac_encode_close,
    .sample_fmts = (const enum SampleFormat[]){SAMPLE_FMT_S16,SAMPLE_FMT_NONE},
    .long_name = NULL_IF_CONFIG_SMALL("libfaac AAC (Advanced Audio Codec)"),
};
